"""
Handlers are a way to describe to Eikobot how something should be deployed.
"""
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Union

from . import logger
from .compiler.definitions.base_model import BaseModel
from .compiler.definitions.base_types import EikoResource
from .errors import EikoUnresolvedPromiseError

CACHE_DIR = Path(".eikobot_cache")
CACHE_DIR.mkdir(exist_ok=True)


@dataclass
class HandlerContext:
    """A HandlerContext keeps track of things required for a deployment."""

    raw_resource: EikoResource
    task_id: str

    def __post_init__(self) -> None:
        self.resource: Union[dict, BaseModel]
        self.changes: dict[str, Any] = {}
        self.deployed = False
        self.updated = False
        self.failed = False
        self.promises = self.raw_resource.promises
        self.name = self.raw_resource.index()
        self.extras: dict[str, Any] = {}
        self.task_cache = CACHE_DIR / self.normalized_task_id()

        self.task_cache.mkdir(exist_ok=True)

    def normalized_task_id(self) -> str:
        """
        Removes backslashes, forward slashes and : from the task_id
        so it can be used for paths on both unix and windows.
        """
        _normalized = self.task_id.replace("\\", "-")
        _normalized = _normalized.replace("/", "-")
        _normalized = _normalized.replace(" ", "")
        return _normalized.replace(":", ".")

    def add_change(self, key: str, value: Any) -> None:
        self.changes[key] = value

    def debug(self, msg: str) -> None:
        logger.debug(f"[{self.name}] {msg}")

    def info(self, msg: str) -> None:
        logger.info(f"[{self.name}] {msg}")

    def warning(self, msg: str) -> None:
        logger.warning(f"[{self.name}] {msg}")

    def error(self, msg: str) -> None:
        logger.error(f"[{self.name}] {msg}")


class Handler:
    """A handler implements methods for a resource to be managed."""

    __eiko_resource__: str

    async def __pre__(self, ctx: HandlerContext) -> None:
        pass

    async def execute(self, ctx: HandlerContext) -> None:
        raise NotImplementedError

    async def resolve_promises(self, ctx: HandlerContext) -> None:
        pass

    async def __post__(self, ctx: HandlerContext) -> None:
        pass

    async def cleanup(self, ctx: HandlerContext) -> None:
        pass

    async def __dry_run__(self, ctx: HandlerContext) -> None:
        ctx.info("Task would execute.")


class EikoCRUDHanlderMethodNotImplemented(Exception):
    """Raised if a method is not implemented."""


class CRUDHandler(Handler):
    """
    An async crud resource handler is like a CRUDHandler,
    but it's methods are async.
    """

    async def execute(self, ctx: HandlerContext) -> None:
        ctx.failed = False
        ctx.deployed = False
        try:
            ctx.debug("Reading resource.")
            await self.read(ctx)
        except EikoCRUDHanlderMethodNotImplemented:
            pass

        if not ctx.deployed:
            try:
                ctx.debug("Deploying resource.")
                await self.create(ctx)
            except EikoCRUDHanlderMethodNotImplemented:
                ctx.error(
                    "Tried to deploy resource, but handler is missing a create method."
                )
                return
        else:
            if ctx.changes:
                try:
                    ctx.deployed = False
                    ctx.debug("Updating resource.")
                    await self.update(ctx)
                except EikoCRUDHanlderMethodNotImplemented:
                    ctx.warning(
                        "Read method returned changes for handler without update method."
                    )
            else:
                ctx.debug("Resource is in its desired state.")

        if not ctx.deployed:
            ctx.failed = True

    async def create(self, ctx: HandlerContext) -> None:
        raise EikoCRUDHanlderMethodNotImplemented

    async def read(self, ctx: HandlerContext) -> None:
        raise EikoCRUDHanlderMethodNotImplemented

    async def update(self, ctx: HandlerContext) -> None:
        raise EikoCRUDHanlderMethodNotImplemented

    async def delete(self, ctx: HandlerContext) -> None:
        raise EikoCRUDHanlderMethodNotImplemented

    async def __dry_run__(self, ctx: HandlerContext) -> None:
        try:
            ctx.debug(f"Reading resource '{ctx.raw_resource.index()}'.")
            await self.read(ctx)
        except EikoCRUDHanlderMethodNotImplemented:
            ctx.info("Resource would be created.")
        except EikoUnresolvedPromiseError:
            ctx.info(
                "Resource relies on promises that are unresolved and thus its state is unknown."
            )
            return

        if ctx.deployed:
            if not ctx.changes:
                ctx.info("Resource is in its desired state.")
            else:
                ctx.info(f"Resource would be updated. (changes: {ctx.changes})")
        elif ctx.failed:
            ctx.error("Resource is in a failed state!")
        else:
            ctx.info("Resource would be created.")



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

import pytest

from eikobot.core.compiler._token import Token, TokenType
from eikobot.core.compiler.lexer import EikoSyntaxError, Lexer
from eikobot.core.compiler.misc import Index


def test_char_generator(eiko_file_1: Path) -> None:
    """
    Tests Lexer._next_char and Lexer._current_index
    """
    i = 0
    line = 0
    col = 0
    lexer = Lexer(eiko_file_1)
    char = lexer._next_char()
    while True:
        if char == "EOF":
            break

        assert lexer._current_index() == Index(line, col, eiko_file_1)
        assert char == lexer._content[i]
        col += 1
        if char == "\n":
            col = 0
            line += 1

        char = lexer._next_char()
        i += 1


def test_token_generator(eiko_file_2: Path) -> None:
    """Tests Lexer._next_char and Lexer._current_index"""
    lexer = Lexer(eiko_file_2)

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(0, 0, eiko_file_2))
    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(1, 0, eiko_file_2))
    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(2, 0, eiko_file_2))

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(3, 0, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.RESOURCE, "resource", Index(3, 0, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "Host", Index(3, 9, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(3, 13, eiko_file_2))

    assert lexer.next_token() == Token(TokenType.INDENT, "\t", Index(4, 0, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "ip", Index(4, 1, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(4, 3, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "IpAddress", Index(4, 5, eiko_file_2)
    )

    assert lexer.next_token() == Token(TokenType.INDENT, "\t", Index(5, 0, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "ssh_key", Index(5, 1, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(5, 8, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "str", Index(5, 10, eiko_file_2)
    )

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(6, 0, eiko_file_2))

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(7, 0, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.RESOURCE, "resource", Index(7, 0, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "Debian", Index(7, 9, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.LEFT_PAREN, "(", Index(7, 15, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "Host", Index(7, 16, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.RIGHT_PAREN, ")", Index(7, 20, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(7, 21, eiko_file_2))

    assert lexer.next_token() == Token(TokenType.INDENT, "\t", Index(8, 0, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "version", Index(8, 1, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(8, 8, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "int", Index(8, 10, eiko_file_2)
    )

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(9, 0, eiko_file_2))

    assert lexer.next_token() == Token(
        TokenType.INDENT, "\t", Index(10, 0, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IMPLEMENT, "implement", Index(10, 1, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "default", Index(10, 11, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.LEFT_PAREN, "(", Index(10, 18, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "self", Index(10, 19, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COMMA, ",", Index(10, 23, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "version", Index(10, 25, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(10, 32, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "int", Index(10, 34, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.RIGHT_PAREN, ")", Index(10, 37, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.COLON, ":", Index(10, 38, eiko_file_2))

    assert lexer.next_token() == Token(
        TokenType.INDENT, "\t\t", Index(11, 0, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "self", Index(11, 2, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.DOT, ".", Index(11, 6, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "version", Index(11, 7, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.ASSIGNMENT_OP, "=", Index(11, 15, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "version", Index(11, 17, eiko_file_2)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "\t\t", Index(12, 0, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "self", Index(12, 2, eiko_file_2)
    )
    assert lexer.next_token() == Token(TokenType.DOT, ".", Index(12, 6, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "software", Index(12, 7, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.ASSIGNMENT_OP, "=", Index(12, 16, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, "nginx", Index(12, 18, eiko_file_2)
    )

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(13, 0, eiko_file_2))

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(14, 0, eiko_file_2))
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "debian_host", Index(14, 0, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.ASSIGNMENT_OP, "=", Index(14, 12, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.IDENTIFIER, "Debian", Index(14, 14, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.LEFT_PAREN, "(", Index(14, 20, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "11", Index(14, 21, eiko_file_2)
    )
    assert lexer.next_token() == Token(
        TokenType.RIGHT_PAREN, ")", Index(14, 23, eiko_file_2)
    )

    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(16, 0, eiko_file_2))

    assert lexer.next_token() == Token(TokenType.EOF, "EOF", Index(16, 0, eiko_file_2))
    # Lexer keeps returning EOF instead of an error. This is by design.
    assert lexer.next_token() == Token(TokenType.EOF, "EOF", Index(16, 0, eiko_file_2))


def test_base_types(eiko_base_type_file: Path) -> None:
    lexer = Lexer(eiko_base_type_file)
    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(0, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING,
        "This test shows off escape \\ sequence parsing.\n",
        Index(0, 0, eiko_base_type_file),
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(1, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "123456", Index(1, 0, eiko_base_type_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(2, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.FLOAT, "123.123", Index(2, 0, eiko_base_type_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(3, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, "base string", Index(3, 0, eiko_base_type_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(4, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, r"\n\n", Index(4, 0, eiko_base_type_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(5, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.F_STRING, "Hello, {name}", Index(5, 0, eiko_base_type_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(7, 0, eiko_base_type_file)
    )
    assert lexer.next_token() == Token(
        TokenType.EOF, "EOF", Index(7, 0, eiko_base_type_file)
    )


def test_basic_math(eiko_basic_ops_file: Path) -> None:
    lexer = Lexer(eiko_basic_ops_file)

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(0, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "1", Index(0, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(0, 2, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "4", Index(0, 4, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "//", Index(0, 6, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(0, 9, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(0, 11, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "3", Index(0, 13, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(1, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "1", Index(1, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(1, 2, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(1, 4, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "//", Index(1, 6, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "-", Index(1, 9, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "3", Index(1, 10, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "**", Index(1, 11, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(1, 13, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(1, 15, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "-", Index(1, 17, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "7", Index(1, 18, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "*", Index(1, 20, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(1, 22, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(2, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.LEFT_PAREN, "(", Index(2, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.LEFT_PAREN, "(", Index(2, 1, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "1", Index(2, 2, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(2, 4, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(2, 6, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.RIGHT_PAREN, ")", Index(2, 7, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "//", Index(2, 9, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "-", Index(2, 12, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "3", Index(2, 13, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.RIGHT_PAREN, ")", Index(2, 14, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "**", Index(2, 15, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(2, 17, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(2, 19, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "-", Index(2, 21, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.LEFT_PAREN, "(", Index(2, 22, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "7", Index(2, 23, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "*", Index(2, 25, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "2", Index(2, 27, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.RIGHT_PAREN, ")", Index(2, 28, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(3, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, "string 1", Index(3, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(3, 11, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, " + ", Index(3, 13, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "+", Index(3, 19, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, "string 2", Index(3, 21, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(4, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "4", Index(4, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "-", Index(4, 2, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "4", Index(4, 4, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "-", Index(4, 6, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "4", Index(4, 8, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(5, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, "ha", Index(5, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.ARITHMETIC_OP, "*", Index(5, 5, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.INTEGER, "3", Index(5, 7, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(6, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, "auto", Index(6, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, " concat", Index(6, 6, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.STRING, " string", Index(6, 15, eiko_basic_ops_file)
    )

    assert lexer.next_token() == Token(
        TokenType.INDENT, "", Index(8, 0, eiko_basic_ops_file)
    )
    assert lexer.next_token() == Token(
        TokenType.EOF, "EOF", Index(8, 0, eiko_basic_ops_file)
    )


def test_comment_end_of_file_lexing(tmp_eiko_file: Path) -> None:
    with open(tmp_eiko_file, "w", encoding="utf-8") as f:
        f.write("# Test comment")

    lexer = Lexer(tmp_eiko_file)
    assert lexer.next_token() == Token(TokenType.INDENT, "", Index(0, 0, tmp_eiko_file))
    assert lexer.next_token() == Token(TokenType.EOF, "EOF", Index(1, 0, tmp_eiko_file))


def test_unicode_escape_error(tmp_eiko_file: Path) -> None:
    file_path = "\\U"
    model = f'path = Path("{file_path}")'
    with open(tmp_eiko_file, "w", encoding="utf-8") as f:
        f.write(model)

    lexer = Lexer(tmp_eiko_file)
    lexer.next_token()
    lexer.next_token()
    lexer.next_token()
    lexer.next_token()
    lexer.next_token()
    with pytest.raises(EikoSyntaxError):
        lexer.next_token()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

from eikobot.core.compiler import Compiler
from eikobot.core.compiler.definitions._resource import EikoResourceDefinition
from eikobot.core.compiler.definitions.base_types import EikoResource
from eikobot.core.compiler.definitions.context import CompilerContext
from eikobot.core.compiler.definitions.function import PluginDefinition
from eikobot.core.compiler.importlib import resolve_import


def test_import_std() -> None:
    context = CompilerContext("__main__", {})
    resolve_import(["std"], context)
    std_context = context.get("std")
    assert isinstance(std_context, CompilerContext)


def test_std(eiko_std_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_std_file)

    test_resource = compiler.context.get("test_resource")
    assert isinstance(test_resource, EikoResource)
    assert test_resource.type.name == "CompileTestResource"

    test_2 = compiler.context.get("test_2")
    assert isinstance(test_2, EikoResource)
    assert test_2.type.name == "Test_2"
    assert test_2.properties.get("test") == test_resource


def test_from_import(eiko_from_import_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_from_import_file)

    test_resource = compiler.context.get("test_resource")
    assert isinstance(test_resource, EikoResource)
    assert test_resource.type.name == "CompileTestResource"

    test_2 = compiler.context.get("test_2")
    assert isinstance(test_2, EikoResource)
    assert test_2.type.name == "Test_2"
    assert test_2.properties.get("test") == test_resource


def test_multi_import(eiko_multi_import_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_multi_import_file)

    std = compiler.context.get("std")
    assert isinstance(std, CompilerContext)

    inspect = compiler.context.get("inspect")
    assert isinstance(inspect, PluginDefinition)

    file = compiler.context.get("file")
    assert isinstance(file, CompilerContext)

    file_definition = compiler.context.get("File")
    assert isinstance(file_definition, EikoResourceDefinition)

    read_file = compiler.context.get("read_file")
    assert isinstance(read_file, PluginDefinition)
